# coding: utf-8
import os
import sys

import pytest

from _pydev_bundle.pydev_monkey import pydev_src_dir
from _pydevd_bundle.pydevd_constants import sorted_dict_repr
from pydevd import SetupHolder

try:
    from _pydev_bundle import pydev_monkey
except:
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))
    from _pydev_bundle import pydev_monkey


@pytest.fixture(autouse=True)
def save_setup_holder():
    original = SetupHolder.setup
    yield
    SetupHolder.setup = original


def test_monkey():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "ppid": os.getpid(), "protocol-quoted-line": True}
    check = """C:\\bin\\python.exe -u -c connect(\\"127.0.0.1\\")"""
    debug_command = (
        "import sys; "
        "sys.path.insert(0, r'%s'); "
        "import pydevd; pydevd.config('quoted-line', ''); "
        "pydevd.settrace(host='127.0.0.1', port=0, suspend=False, "
        "trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); "
        ""
        'connect("127.0.0.1")'
    ) % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command

    assert "C:\\bin\\python.exe -u -c %s" % debug_command == pydev_monkey.patch_arg_str_win(check)


def test_str_to_args_windows():
    assert ["a", "b"] == pydev_monkey.str_to_args_windows('a "b"')


def test_monkey_patch_return_original_args():
    check = ["echo", '"my"', '"args"']
    res = pydev_monkey.patch_args(check[:])
    assert res == check


def test_monkey_patch_pathlib_args():
    try:
        import pathlib
    except ImportError:
        pytest.skip("pathlib not available.")

    check = [pathlib.Path("echo"), '"my"', '"args"']
    res = pydev_monkey.patch_args(check[:])
    assert res == check


def test_monkey_patch_wrong_object_type():
    check = [1, 22, '"my"', '"args"']
    res = pydev_monkey.patch_args(check[:])
    assert res == check


def test_monkey_patch_wrong_object_type_2():
    check = ["C:\\bin\\python.exe", "-u", 1, '-qcconnect("127.0.0.1")']
    res = pydev_monkey.patch_args(check[:])
    assert res == check


def test_monkey_patch_args_module_subprocess_pathlib():
    try:
        import pathlib
    except ImportError:
        pytest.skip("pathlib not available.")

    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "multiprocess": True}
    if sys.platform == "win32":
        python_path = "C:\\bin\\python.exe"
    else:
        python_path = "/bin/python"
    check = [pathlib.Path(python_path), "-mtest", pathlib.Path("bar")]
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file

    assert pydev_monkey.patch_args(check) == [
        python_path,
        get_pydevd_file(),
        "--module",
        "--port",
        "0",
        "--ppid",
        str(os.getpid()),
        "--client",
        "127.0.0.1",
        "--multiprocess",
        "--protocol-quoted-line",
        "--file",
        "test",
        "bar",
    ]


def test_monkey_patch_args_indc():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "ppid": os.getpid(), "protocol-quoted-line": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-u", "-c", 'connect("127.0.0.1")']
    debug_command = (
        "import sys; sys.path.insert(0, r'%s'); import pydevd; pydevd.config('quoted-line', ''); "
        "pydevd.settrace(host='127.0.0.1', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); "
        ""
        'connect("127.0.0.1")'
    ) % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == ["C:\\bin\\python.exe", "-u", "-c", debug_command]


def test_separate_future_imports():
    found = pydev_monkey._separate_future_imports("""from __future__ import print_function\nprint(1)""")
    assert found == ("from __future__ import print_function;", "\nprint(1)")

    found = pydev_monkey._separate_future_imports("""from __future__ import print_function;print(1)""")
    assert found == ("from __future__ import print_function;", "print(1)")

    found = pydev_monkey._separate_future_imports("""from __future__ import (\nprint_function);print(1)""")
    assert found == ("from __future__ import (\nprint_function);", "print(1)")

    found = pydev_monkey._separate_future_imports(""""line";from __future__ import (\n\nprint_function, absolute_import\n);print(1)""")
    assert found == ('"line";from __future__ import (\n\nprint_function, absolute_import\n);', "print(1)")

    found = pydev_monkey._separate_future_imports(
        """from __future__ import division\nfrom __future__ import (\n\nprint_function, absolute_import\n);print(1)"""
    )
    assert found == ("from __future__ import division\nfrom __future__ import (\n\nprint_function, absolute_import\n);", "print(1)")


def test_monkey_patch_args_indc_future_import():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "ppid": os.getpid(), "protocol-quoted-line": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-u", "-c", 'from __future__ import print_function;connect("127.0.0.1")']
    debug_command = (
        "from __future__ import print_function;import sys; sys.path.insert(0, r'%s'); import pydevd; pydevd.config('quoted-line', ''); "
        "pydevd.settrace(host='127.0.0.1', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); "
        ""
        'connect("127.0.0.1")'
    ) % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == ["C:\\bin\\python.exe", "-u", "-c", debug_command]


def test_monkey_patch_args_indc_future_import2():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "ppid": os.getpid(), "protocol-quoted-line": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-u", "-c", 'from __future__ import print_function\nconnect("127.0.0.1")']
    debug_command = (
        "from __future__ import print_function;import sys; sys.path.insert(0, r'%s'); import pydevd; pydevd.config('quoted-line', ''); "
        "pydevd.settrace(host='127.0.0.1', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); "
        ""
        '\nconnect("127.0.0.1")'
    ) % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == ["C:\\bin\\python.exe", "-u", "-c", debug_command]


def test_monkey_patch_args_indc2():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "ppid": os.getpid(), "protocol-quoted-line": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-u", '-qcconnect("127.0.0.1")']
    debug_command = (
        "import sys; sys.path.insert(0, r'%s'); import pydevd; pydevd.config('quoted-line', ''); "
        "pydevd.settrace(host='127.0.0.1', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); "
        ""
        'connect("127.0.0.1")'
    ) % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == ["C:\\bin\\python.exe", "-u", "-qc", debug_command]


def test_monkey_patch_args_x_flag():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "ppid": os.getpid(), "protocol-quoted-line": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-X", "faulthandler", "-c", 'connect("127.0.0.1")']
    debug_command = (
        "import sys; sys.path.insert(0, r'%s'); import pydevd; pydevd.config('quoted-line', ''); "
        "pydevd.settrace(host='127.0.0.1', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); "
        ""
        'connect("127.0.0.1")'
    ) % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == ["C:\\bin\\python.exe", "-X", "faulthandler", "-c", debug_command]


def test_monkey_patch_args_flag_in_single_arg_1():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "ppid": os.getpid(), "protocol-quoted-line": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-qX", "faulthandler", "-c", 'connect("127.0.0.1")']
    debug_command = (
        "import sys; sys.path.insert(0, r'%s'); import pydevd; pydevd.config('quoted-line', ''); "
        "pydevd.settrace(host='127.0.0.1', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); "
        ""
        'connect("127.0.0.1")'
    ) % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == ["C:\\bin\\python.exe", "-qX", "faulthandler", "-c", debug_command]


def test_monkey_patch_args_flag_in_single_arg_2():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "ppid": os.getpid(), "protocol-quoted-line": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-qX", "faulthandler", "-c", 'connect("127.0.0.1")']
    debug_command = (
        "import sys; sys.path.insert(0, r'%s'); import pydevd; pydevd.config('quoted-line', ''); "
        "pydevd.settrace(host='127.0.0.1', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); "
        ""
        'connect("127.0.0.1")'
    ) % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == ["C:\\bin\\python.exe", "-qX", "faulthandler", "-c", debug_command]


def test_monkey_patch_args_flag_in_single_arg_3():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "ppid": os.getpid(), "protocol-quoted-line": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-qc", 'connect("127.0.0.1")']
    debug_command = (
        "import sys; sys.path.insert(0, r'%s'); import pydevd; pydevd.config('quoted-line', ''); "
        "pydevd.settrace(host='127.0.0.1', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); "
        ""
        'connect("127.0.0.1")'
    ) % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == ["C:\\bin\\python.exe", "-qc", debug_command]


def test_monkey_patch_args_x_flag_inline():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "ppid": os.getpid(), "protocol-quoted-line": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-Xfaulthandler", "-c", 'connect("127.0.0.1")', "arg1"]
    debug_command = (
        "import sys; sys.path.insert(0, r'%s'); import pydevd; pydevd.config('quoted-line', ''); "
        "pydevd.settrace(host='127.0.0.1', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); "
        ""
        'connect("127.0.0.1")'
    ) % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == ["C:\\bin\\python.exe", "-Xfaulthandler", "-c", debug_command, "arg1"]


def test_monkey_patch_args_c_flag_inline():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "ppid": os.getpid(), "protocol-quoted-line": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-X", "faulthandler", '-cconnect("127.0.0.1")', "arg1"]
    debug_command = (
        "import sys; sys.path.insert(0, r'%s'); import pydevd; pydevd.config('quoted-line', ''); "
        "pydevd.settrace(host='127.0.0.1', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); "
        ""
        'connect("127.0.0.1")'
    ) % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == ["C:\\bin\\python.exe", "-X", "faulthandler", "-c", debug_command, "arg1"]


def test_monkey_patch_args_module():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "multiprocess": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-m", "test"]
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file

    assert pydev_monkey.patch_args(check) == [
        "C:\\bin\\python.exe",
        get_pydevd_file(),
        "--module",
        "--port",
        "0",
        "--ppid",
        str(os.getpid()),
        "--client",
        "127.0.0.1",
        "--multiprocess",
        "--skip-notify-stdin",
        "--protocol-quoted-line",
        "--file",
        "test",
    ]


def test_monkey_patch_args_unbuffered_module():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "multiprocess": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-u", "-m", "test"]
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file

    assert pydev_monkey.patch_args(check) == [
        "C:\\bin\\python.exe",
        "-u",
        get_pydevd_file(),
        "--module",
        "--port",
        "0",
        "--ppid",
        str(os.getpid()),
        "--client",
        "127.0.0.1",
        "--multiprocess",
        "--skip-notify-stdin",
        "--protocol-quoted-line",
        "--file",
        "test",
    ]


def test_monkey_patch_args_module_inline():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "multiprocess": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-qOmtest"]
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file

    assert pydev_monkey.patch_args(check) == [
        "C:\\bin\\python.exe",
        "-qO",
        get_pydevd_file(),
        "--module",
        "--port",
        "0",
        "--ppid",
        str(os.getpid()),
        "--client",
        "127.0.0.1",
        "--multiprocess",
        "--skip-notify-stdin",
        "--protocol-quoted-line",
        "--file",
        "test",
    ]


def test_monkey_patch_args_module_inline2():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "multiprocess": True, "skip-notify-stdin": True}
    check = ["C:\\bin\\python.exe", "-qOm", "test"]
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file

    assert pydev_monkey.patch_args(check) == [
        "C:\\bin\\python.exe",
        "-qO",
        get_pydevd_file(),
        "--module",
        "--port",
        "0",
        "--ppid",
        str(os.getpid()),
        "--client",
        "127.0.0.1",
        "--multiprocess",
        "--skip-notify-stdin",
        "--protocol-quoted-line",
        "--file",
        "test",
    ]


def test_monkey_patch_args_no_indc():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0"}
    check = ["C:\\bin\\python.exe", 'connect(\\"127.0.0.1\\")', "with spaces"]
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file

    assert pydev_monkey.patch_args(check) == [
        "C:\\bin\\python.exe",
        get_pydevd_file(),
        "--port",
        "0",
        "--ppid",
        str(os.getpid()),
        "--client",
        "127.0.0.1",
        "--protocol-quoted-line",
        "--file",
        '"connect(\\\\\\"127.0.0.1\\\\\\")"' if sys.platform == "win32" else 'connect(\\"127.0.0.1\\")',
        '"with spaces"' if sys.platform == "win32" else "with spaces",
    ]


def test_monkey_patch_args_no_indc_with_pydevd():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0"}
    check = ["C:\\bin\\python.exe", "pydevd.py", 'connect(\\"127.0.0.1\\")', "bar"]

    assert pydev_monkey.patch_args(check) == ["C:\\bin\\python.exe", "pydevd.py", 'connect(\\"127.0.0.1\\")', "bar"]


def test_monkey_patch_args_no_indc_without_pydevd():
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file

    SetupHolder.setup = {"client": "127.0.0.1", "port": "0"}
    check = ["C:\\bin\\python.exe", "target.py", 'connect(\\"127.0.0.1\\")', "bar"]
    assert pydev_monkey.patch_args(check) == [
        "C:\\bin\\python.exe",
        get_pydevd_file(),
        "--port",
        "0",
        "--ppid",
        str(os.getpid()),
        "--client",
        "127.0.0.1",
        "--protocol-quoted-line",
        "--file",
        "target.py",
        '"connect(\\\\\\"127.0.0.1\\\\\\")"' if sys.platform == "win32" else 'connect(\\"127.0.0.1\\")',
        "bar",
    ]


@pytest.mark.parametrize("use_bytes", [True, False])
def test_monkey_patch_c_program_arg(use_bytes):
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file

    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "module": "ignore.this"}
    check = ["C:\\bin\\python.exe", "-u", "target.py", "-c", "-áéíóú"]

    encode = lambda s: s
    if use_bytes:
        check = [c.encode("utf-8") for c in check]
        encode = lambda s: s.encode("utf-8")

    assert pydev_monkey.patch_args(check) == [
        encode("C:\\bin\\python.exe"),
        encode("-u"),
        get_pydevd_file(),
        "--port",
        "0",
        "--ppid",
        str(os.getpid()),
        "--client",
        "127.0.0.1",
        "--protocol-quoted-line",
        "--file",
        encode("target.py"),
        encode("-c"),
        encode("-áéíóú"),
    ]


def test_monkey_patch_args_module_single_arg():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "multiprocess": True, "module": "ignore.this"}
    check = ["C:\\bin\\python.exe", "-mtest", "bar"]
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file

    assert pydev_monkey.patch_args(check) == [
        "C:\\bin\\python.exe",
        get_pydevd_file(),
        "--module",
        "--port",
        "0",
        "--ppid",
        str(os.getpid()),
        "--client",
        "127.0.0.1",
        "--multiprocess",
        "--protocol-quoted-line",
        "--file",
        "test",
        "bar",
    ]


def test_monkey_patch_args_stdin():
    SetupHolder.setup = {"client": "127.0.0.1", "port": "0", "multiprocess": True, "module": "ignore.this"}
    check = ["C:\\bin\\python.exe", "-Xfaulthandler", "-"]
    # i.e.: we don't deal with the stdin.
    assert pydev_monkey.patch_args(check) == check



from _pydevd_bundle import pydevd_utils
from _pydevd_bundle.pydevd_additional_thread_info import set_additional_thread_info
from _pydevd_bundle.pydevd_comm_constants import CMD_STEP_INTO, CMD_THREAD_SUSPEND
from _pydevd_bundle.pydevd_constants import PYTHON_SUSPEND, STATE_SUSPEND, get_thread_id, STATE_RUN, PYDEVD_USE_SYS_MONITORING
from _pydev_bundle._pydev_saved_modules import threading
from _pydev_bundle import pydev_log
import sys
from _pydevd_sys_monitoring import pydevd_sys_monitoring


def pydevd_find_thread_by_id(thread_id):
    try:
        threads = threading.enumerate()
        for i in threads:
            tid = get_thread_id(i)
            if thread_id == tid or thread_id.endswith("|" + tid):
                return i

        # This can happen when a request comes for a thread which was previously removed.
        pydev_log.info("Could not find thread %s.", thread_id)
        pydev_log.info("Available: %s.", ([get_thread_id(t) for t in threads],))
    except:
        pydev_log.exception()

    return None


def mark_thread_suspended(thread, stop_reason: int, original_step_cmd: int = -1, main_suspend: bool = True):
    pydev_log.info("Marking thread suspended. Name: %s, stop_reason: %s, main_suspend: %s", thread.name, stop_reason, main_suspend)
    info = set_additional_thread_info(thread)
    info.suspend_type = PYTHON_SUSPEND
    if original_step_cmd != -1:
        stop_reason = original_step_cmd
    thread.stop_reason = stop_reason

    # Note: don't set the 'pydev_original_step_cmd' here if unset.

    if not main_suspend:
        info.pydev_step_cmd = CMD_THREAD_SUSPEND
        info.pydev_step_stop = None
    elif info.pydev_step_cmd == -1:
        # If the step command is not specified, set it to step into
        # to make sure it'll break as soon as possible.
        info.pydev_step_cmd = CMD_STEP_INTO
        info.pydev_step_stop = None

    # Mark as suspended as the last thing.
    info.pydev_state = STATE_SUSPEND
    info.update_stepping_info()
    return info


def internal_run_thread(thread, set_additional_thread_info):
    info = set_additional_thread_info(thread)
    info.pydev_original_step_cmd = -1
    info.pydev_step_cmd = -1
    info.pydev_step_stop = None
    info.pydev_state = STATE_RUN
    info.update_stepping_info()


def resume_threads(thread_id, except_thread=None):
    pydev_log.info("Resuming threads: %s (except thread: %s)", thread_id, except_thread)
    threads = []
    if thread_id == "*":
        threads = pydevd_utils.get_non_pydevd_threads()

    elif thread_id.startswith("__frame__:"):
        pydev_log.critical("Can't make tasklet run: %s", thread_id)

    else:
        threads = [pydevd_find_thread_by_id(thread_id)]

    for t in threads:
        if t is None or t is except_thread:
            pydev_log.info("Skipped resuming thread: %s", t)
            continue

        internal_run_thread(t, set_additional_thread_info=set_additional_thread_info)


from _pydevd_bundle.pydevd_constants import ForkSafeLock

suspend_threads_lock = ForkSafeLock()


def suspend_all_threads(py_db, except_thread):
    """
    Suspend all except the one passed as a parameter.
    :param except_thread:
    """
    if PYDEVD_USE_SYS_MONITORING:
        pydevd_sys_monitoring.update_monitor_events(suspend_requested=True)

    pydev_log.info("Suspending all threads except: %s", except_thread)
    all_threads = pydevd_utils.get_non_pydevd_threads()
    for t in all_threads:
        if getattr(t, "pydev_do_not_trace", None):
            pass  # skip some other threads, i.e. ipython history saving thread from debug console
        else:
            if t is except_thread:
                continue
            info = mark_thread_suspended(t, CMD_THREAD_SUSPEND, main_suspend=False)
            frame = info.get_topmost_frame(t)

            # Reset the tracing as in this case as it could've set scopes to be untraced.
            if frame is not None:
                try:
                    py_db.set_trace_for_frame_and_parents(t.ident, frame)
                finally:
                    frame = None

    if PYDEVD_USE_SYS_MONITORING:
        # After suspending the frames we need the monitoring to be reset.
        pydevd_sys_monitoring.restart_events()
